==> 0.0.2.tar.gz: lib/pages/compose.dart <==
import 'dart:async';
import 'dart:io';

import 'package:badges/badges.dart';
import 'package:bizdoc/core/app.dart';
import 'package:file_picker/file_picker.dart';
import 'package:flutter/material.dart';
import 'package:flutter/scheduler.dart';
import 'package:provider/provider.dart';

import 'package:bizdoc/gen_l10n/app_localizations.dart';
import 'package:bizdoc/core/extensions.dart';
import 'package:bizdoc/core/form.dart';
import 'attachments.dart';
import 'comments.dart';
import 'fields.dart';
import 'package:bizdoc/models/contracts.dart';
import 'package:bizdoc/models/models.dart';
import 'package:bizdoc/models/profile.dart';
import 'package:bizdoc/core/hub.dart';
import 'package:bizdoc/services/accounts.dart';
import 'package:bizdoc/services/base.dart';
import 'package:bizdoc/services/mailbox.dart';
import 'package:bizdoc/services/session.dart';

/// Detail of [RecipientModel]. Enable save, submit, send, attach.
class ComposePage extends StatefulWidget {
  static String routeName = 'mailbox/i';

  const ComposePage({Key? key}) : super(key: key);

  @override
  _ComposePageState createState() => _ComposePageState();
}

class ComposeModel {
  final RecipientModel model;
  final String? action;
  ComposeModel(this.model, {this.action});
  factory ComposeModel.fromJson(Map<String, dynamic> json) {
    return ComposeModel(RecipientModel.fromJson(json['model']),
        action: json['action']);
  }
}

class _ComposePageState extends State<ComposePage> {
  bool _processing = false;
  late RecipientModel _model;
  late FormInfo _form;
  late BizDocFormState _composeState;

  Widget? form;

  @override
  void initState() {
    super.initState();
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final model = ModalRoute.of(context)!.settings.arguments as ComposeModel;
    _model = model.model;
    _form = session.profile.forms
        .firstWhere((element) => element.name == _model.form);
    _composeState = BizDocFormState(_model, ViewMode.compose);
    _listen();
  }

  ///
  Future<void> _listen() async {
    await for (var contract in hub.events.where((event) =>
        (event is CommentContractBase && event.id == _model.documentId) ||
        (event is MailUpdate && event.model.id == _model.id))) {
      if (contract is NewComment) {
        try {
          final info = await accounts.userInfo(contract.model.userId);
          contract.model.name = info.name;
          _model.comments.add(contract.model);
          if (!Scaffold.of(context).isEndDrawerOpen) {
            setState(() {
              _model.unreadComments++;
            });
          }
        } catch (e) {
          print(e);
        }
      } else if (contract is CommentUpdate) {
        final comment = _model.comments
            .firstWhere((element) => element.id == contract.model.id);
        comment
          ..text = contract.model.text
          ..edited = contract.model.edited;
      } else if (contract is CommentDelete) {
        _model.comments
            .removeWhere((element) => element.id == contract.commentId);
      } else if (contract is MailUpdate) {
        _model.update(contract.model);
        await _notifyChanged(contract);
      }
    }
    await for (var _ in _composeState.state) {
      setState(() {});
    }
  }

  Future<void> _notifyChanged(MailUpdate contract) async {
    try {
      final info = await accounts.userInfo(contract.userId);
      ScaffoldMessenger.of(context).showSnackBar(SnackBar(
        content:
            Text(AppLocalizations.of(context)!.modifiedBy.format([info.name])),
        action: SnackBarAction(
          label: AppLocalizations.of(context)!.ok,
          onPressed: () => setState(() {}),
        ),
      ));
    } catch (e) {}
  }

  @override
  Widget build(BuildContext context) {
    return Consumer<AppState>(builder: (context, value, child) {
      final form = value.resolve(_form.template, context);
      return MultiProvider(
          providers: [
            ChangeNotifierProvider<BizDocFormState>.value(value: _composeState)
          ],
          child: Scaffold(
              appBar: AppBar(
                title: Text('#${_model.number}'),
                actions: <Widget>[
                  if (_model.draft)
                    IconButton(
                        enableFeedback: !_processing,
                        onPressed: _discard,
                        icon: const Icon(Icons.delete)),
                  IconButton(
                      onPressed: _upload, icon: const Icon(Icons.photo_album)),
                  if (_model.actions!.length > 1)
                    PopupMenuButton<String>(
                        icon: const Icon(Icons.send),
                        enableFeedback: !_processing,
                        onSelected: (value) => _onSend(value),
                        itemBuilder: (context) {
                          var actions = session.profile.actions.where(
                              (element) =>
                                  _model.actions!.contains(element.name));
                          final entries = <PopupMenuEntry<String>>[];
                          int? group = actions.first.group;
                          for (var action in actions) {
                            if (action.group != group) {
                              group = action.group;
                              entries.add(const PopupMenuDivider());
                            }
                            entries.add(PopupMenuItem(
                              child: Text(action.title),
                              value: action.name,
                            ));
                          }
                          return entries;
                        }),
                  if (_model.pending || _model.draft)
                    StreamBuilder<bool>(
                      stream: _composeState.dirty,
                      builder: (context, snapshot) => IconButton(
                          onPressed: _onSave,
                          enableFeedback:
                              _processing && snapshot.hasData && snapshot.data!,
                          icon: const Icon(Icons.save)),
                    ),
                  if (_model.draft)
                    StreamBuilder<bool>(
                        stream: _composeState.valid,
                        builder: (context, snapshot) => IconButton(
                            enableFeedback: snapshot.hasData && snapshot.data!,
                            onPressed: _onSubmit,
                            icon: const Icon(Icons.send))),
                  Builder(
                    builder: (context) => Badge(
                      showBadge: _model.unreadComments > 0,
                      badgeContent: Text(_model.unreadComments.toString()),
                      position: const BadgePosition(top: 0),
                      child: IconButton(
                          onPressed: () {
                            Scaffold.of(context).openEndDrawer();
                            _commentsSeen();
                          },
                          icon: const Icon(Icons.comment)),
                    ),
                  ),
                ],
              ),
              endDrawer: Drawer(
                child: CommentsWidget(
                  model: _model,
                  onTap: () => _commentsSeen(),
                ),
              ),
              body: Column(
                children: [
                  if (_processing) const LinearProgressIndicator(),
                  AttachmentsWidget(_model),
                  Expanded(
                    child: form!,
                  )
                ],
              )));
    });
  }

  Future<void> _commentsSeen() async {
    if (_model.viewed == null ||
        _model.comments.any((element) =>
            element.time.isAfter(_model.viewed!) &&
            element.userId !=
                (session.profile.byId ?? session.profile.userId))) {
      try {
        final time = await mailbox.seenComments(_model.documentId);
        _model.viewed = time;
        setState(() {
          _model.unreadComments = 0;
        });
      } catch (e) {}
    }
  }

  void _onSend(String action) async {
    final aaction =
        session.profile.actions.firstWhere((element) => element.name == action);
    final args = await ActionDialog.show(context, aaction);
    if (args != null) {
      _toggleProcessing(true);
      try {
        await mailbox.send(
            _model.id, _model.version, action, args, _model.data);
        final aaction = session.profile.actions
            .firstWhere((element) => element.name == action);
        _snackBar(AppLocalizations.of(context)!.itemAction.format([
          _model.number,
          aaction.adjective ?? aaction.past ?? aaction.title
        ]));
        SchedulerBinding.instance!.addPostFrameCallback((_) {
          Navigator.pop(context);
        });
      } on ConflictError catch (e) {
        _conflict(e);
      } on ValidationError catch (e) {
        _snackBar(e.message);
        _toggleProcessing(false);
      } catch (e) {
        _snackBar(AppLocalizations.of(context)!.err);
        _toggleProcessing(false);
      }
    }
  }

  void _onSubmit() async {
    _toggleProcessing(true);
    try {
      await mailbox.submit(_model.id, _model.version, _model.data);
      SchedulerBinding.instance!.addPostFrameCallback((_) {
        Navigator.pop(context);
      });
      _snackBar(
          AppLocalizations.of(context)!.submitted.format([_model.number]));
    } on ConflictError catch (e) {
      _conflict(e);
    } on ValidationError catch (e) {
      _snackBar(e.message);
      _toggleProcessing(false);
    } catch (e) {
      _snackBar(AppLocalizations.of(context)!.err);
      _toggleProcessing(false);
    }
  }

  void _snackBar(String message) =>
      ScaffoldMessenger.of(context).showSnackBar(SnackBar(
        content: Text(message),
      ));

  void _onSave() async {
    _toggleProcessing(true);
    try {
      final model = await mailbox.save(_model.id, _model.version, _model.data);
      _model.update(model);
      _composeState.setPristine();
      _snackBar(AppLocalizations.of(context)!.saved);
    } on ConflictError catch (e) {
      _conflict(e);
    } on ValidationError catch (e) {
      _snackBar(e.message);
    } catch (e) {
      _snackBar(AppLocalizations.of(context)!.err);
    } finally {
      _toggleProcessing(false);
    }
  }

  void _conflict(ConflictError e) {
    ScaffoldMessenger.of(context).showSnackBar(SnackBar(
      content: Text(AppLocalizations.of(context)!.err),
      action: SnackBarAction(
        label: AppLocalizations.of(context)!.ok,
        onPressed: () {
          _model.update(e.data);
          _composeState.reset();
        },
      ),
    ));
    _toggleProcessing(false);
  }

  void _toggleProcessing(bool val) => setState(() => _processing = val);

  ///
  void _upload() async {
    try {
      final result = await FilePicker.platform.pickFiles(allowMultiple: true);
      _toggleProcessing(true);
      if (result != null) {
        final files = result.paths.map((path) => File(path!));
        final model = await mailbox.upload(_model.id, _model.version, files);
        _model.attachments.addAll(model);
        _composeState.setPristine();
        setState(() {});
      }
    } catch (e) {
      _snackBar(AppLocalizations.of(context)!.err);
    } finally {
      _toggleProcessing(false);
    }
  }

  void _discard() async {
    final ok = await showDialog<bool?>(
      context: context,
      builder: (context) => AlertDialog(
        content: Text(AppLocalizations.of(context)!.discardAsk),
        actions: [
          TextButton(
            child: Text(AppLocalizations.of(context)!.cancel),
            onPressed: () => Navigator.pop(context),
          ),
          TextButton(
            child: Text(AppLocalizations.of(context)!.discard),
            onPressed: () => Navigator.pop(context, true),
          )
        ],
      ),
    );
    if (ok != true) return;
    _toggleProcessing(true);
    try {
      await mailbox.discard(_model.id);
      SchedulerBinding.instance!.addPostFrameCallback((_) {
        Navigator.pop(context);
      });
    } catch (e) {
      _snackBar(AppLocalizations.of(context)!.err);
      _toggleProcessing(false);
    }
  }

  @override
  void dispose() {
    _composeState.dispose();
    super.dispose();
  }
}

class ActionDialog {
  static Future<Map<String, dynamic>?> show(
      BuildContext context, ActionInfo action) {
    var args = <String, dynamic>{};
    return showDialog<Map<String, dynamic>>(
      context: context,
      builder: (context) {
        return AlertDialog(
          title: Text(action.title),
          content: Fields(
              fields: action.arguments,
              onChanged: (values) => args.addAll(values)),
          actions: [
            TextButton(
              child: Text(AppLocalizations.of(context)!.cancel),
              onPressed: () => Navigator.pop(context),
            ),
            TextButton(
              child: Text(AppLocalizations.of(context)!.oK),
              onPressed: () => Navigator.pop(context, args),
            )
          ],
        );
      },
    );
  }
}

==> 0.0.3.tar.gz: lib/pages/compose.dart <==
import 'dart:async';
import 'dart:io';

import 'package:badges/badges.dart';
import 'package:bizdoc/core/app.dart';
import 'package:file_picker/file_picker.dart';
import 'package:flutter/material.dart';
import 'package:flutter/scheduler.dart';
import 'package:provider/provider.dart';

import 'package:bizdoc/gen_l10n/app_localizations.dart';
import 'package:bizdoc/core/extensions.dart';
import 'package:bizdoc/core/form.dart';
import 'attachments.dart';
import 'comments.dart';
import 'fields.dart';
import 'package:bizdoc/models/contracts.dart';
import 'package:bizdoc/models/data.dart';
import 'package:bizdoc/models/profile.dart';
import 'package:bizdoc/core/hub.dart';
import 'package:bizdoc/services/accounts.dart';
import 'package:bizdoc/services/base.dart';
import 'package:bizdoc/services/mailbox.dart';
import 'package:bizdoc/services/session.dart';

/// Detail of [RecipientModel]. Enable save, submit, send, attach.
class ComposePage extends StatefulWidget {
  static String routeName = 'mailbox/i';

  const ComposePage({Key? key}) : super(key: key);

  @override
  _ComposePageState createState() => _ComposePageState();
}

class ComposeModel {
  final RecipientModel model;
  final String? action;
  ComposeModel(this.model, {this.action});
  factory ComposeModel.fromJson(Map<String, dynamic> json) {
    return ComposeModel(RecipientModel.fromJson(json['model']),
        action: json['action']);
  }
}

class _ComposePageState extends State<ComposePage> {
  bool _processing = false;
  late RecipientModel _model;
  late FormInfo _form;
  late BizDocFormState _composeState;

  Widget? form;

  @override
  void initState() {
    super.initState();
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final model = ModalRoute.of(context)!.settings.arguments as ComposeModel;
    _model = model.model;
    _form = session.profile.forms
        .firstWhere((element) => element.name == _model.form);
    _composeState = BizDocFormState(_model, ViewMode.compose);
    _listen();
  }

  ///
  Future<void> _listen() async {
    await for (var contract in hub.events.where((event) =>
        (event is CommentContractBase && event.id == _model.documentId) ||
        (event is MailUpdate && event.model.id == _model.id))) {
      if (contract is NewComment) {
        try {
          final info = await accounts.find(contract.model.userId);
          contract.model.name = info.name;
          _model.comments.add(contract.model);
          if (!Scaffold.of(context).isEndDrawerOpen) {
            setState(() {
              _model.unreadComments++;
            });
          }
        } catch (e) {
          print(e);
        }
      } else if (contract is CommentUpdate) {
        final comment = _model.comments
            .firstWhere((element) => element.id == contract.model.id);
        comment
          ..text = contract.model.text
          ..edited = contract.model.edited;
      } else if (contract is CommentDelete) {
        _model.comments
            .removeWhere((element) => element.id == contract.commentId);
      } else if (contract is MailUpdate) {
        _model.update(contract.model);
        await _notifyChanged(contract);
      }
    }
    await for (var _ in _composeState.state) {
      setState(() {});
    }
  }

  Future<void> _notifyChanged(MailUpdate contract) async {
    try {
      final info = await accounts.find(contract.userId);
      ScaffoldMessenger.of(context).showSnackBar(SnackBar(
        content:
            Text(AppLocalizations.of(context)!.modifiedBy.format([info.name])),
        action: SnackBarAction(
          label: AppLocalizations.of(context)!.ok,
          onPressed: () => setState(() {}),
        ),
      ));
    } catch (e) {}
  }

  @override
  Widget build(BuildContext context) {
    return Consumer<AppState>(builder: (context, value, child) {
      final form = value.resolve(_form.template, context);
      return MultiProvider(
          providers: [
            ChangeNotifierProvider<BizDocFormState>.value(value: _composeState)
          ],
          child: Scaffold(
              appBar: AppBar(
                title: Text('#${_model.number}'),
                actions: <Widget>[
                  if (_model.draft)
                    IconButton(
                        enableFeedback: !_processing,
                        onPressed: _discard,
                        icon: const Icon(Icons.delete)),
                  IconButton(
                      onPressed: _upload, icon: const Icon(Icons.photo_album)),
                  if (_model.actions!.length > 1)
                    PopupMenuButton<String>(
                        icon: const Icon(Icons.send),
                        enableFeedback: !_processing,
                        onSelected: (value) => _onSend(value),
                        itemBuilder: (context) {
                          var actions = session.profile.actions.where(
                              (element) =>
                                  _model.actions!.contains(element.name));
                          final entries = <PopupMenuEntry<String>>[];
                          int? group = actions.first.group;
                          for (var action in actions) {
                            if (action.group != group) {
                              group = action.group;
                              entries.add(const PopupMenuDivider());
                            }
                            entries.add(PopupMenuItem(
                              child: Text(action.title),
                              value: action.name,
                            ));
                          }
                          return entries;
                        }),
                  if (_model.pending || _model.draft)
                    StreamBuilder<bool>(
                      stream: _composeState.dirty,
                      builder: (context, snapshot) => IconButton(
                          onPressed: _onSave,
                          enableFeedback:
                              _processing && snapshot.hasData && snapshot.data!,
                          icon: const Icon(Icons.save)),
                    ),
                  if (_model.draft)
                    StreamBuilder<bool>(
                        stream: _composeState.valid,
                        builder: (context, snapshot) => IconButton(
                            enableFeedback: snapshot.hasData && snapshot.data!,
                            onPressed: _onSubmit,
                            icon: const Icon(Icons.send))),
                  Builder(
                    builder: (context) => Badge(
                      showBadge: _model.unreadComments > 0,
                      badgeContent: Text(_model.unreadComments.toString()),
                      position: const BadgePosition(top: 0),
                      child: IconButton(
                          onPressed: () {
                            Scaffold.of(context).openEndDrawer();
                            _commentsSeen();
                          },
                          icon: const Icon(Icons.comment)),
                    ),
                  ),
                ],
              ),
              endDrawer: Drawer(
                child: CommentsWidget(
                  model: _model,
                  onTap: () => _commentsSeen(),
                ),
              ),
              body: Column(
                children: [
                  if (_processing) const LinearProgressIndicator(),
                  AttachmentsWidget(_model),
                  Expanded(
                    child: form!,
                  )
                ],
              )));
    });
  }

  Future<void> _commentsSeen() async {
    if (_model.viewed == null ||
        _model.comments.any((element) =>
            element.time.isAfter(_model.viewed!) &&
            element.userId !=
                (session.profile.byId ?? session.profile.userId))) {
      try {
        final time = await mailbox.seenComments(_model.documentId);
        _model.viewed = time;
        setState(() {
          _model.unreadComments = 0;
        });
      } catch (e) {}
    }
  }

  void _onSend(String action) async {
    final aaction =
        session.profile.actions.firstWhere((element) => element.name == action);
    final args = await ActionDialog.show(context, aaction);
    if (args != null) {
      _toggleProcessing(true);
      try {
        await mailbox.send(
            _model.id, _model.version, action, args, _model.data);
        final aaction = session.profile.actions
            .firstWhere((element) => element.name == action);
        _snackBar(AppLocalizations.of(context)!.itemAction.format([
          _model.number,
          aaction.adjective ?? aaction.past ?? aaction.title
        ]));
        SchedulerBinding.instance!.addPostFrameCallback((_) {
          Navigator.pop(context);
        });
      } on ConflictError catch (e) {
        _conflict(e);
      } on ValidationError catch (e) {
        _snackBar(e.message);
        _toggleProcessing(false);
      } catch (e) {
        _snackBar(AppLocalizations.of(context)!.err);
        _toggleProcessing(false);
      }
    }
  }

  void _onSubmit() async {
    _toggleProcessing(true);
    try {
      await mailbox.submit(_model.id, _model.version, _model.data);
      SchedulerBinding.instance!.addPostFrameCallback((_) {
        Navigator.pop(context);
      });
      _snackBar(
          AppLocalizations.of(context)!.submitted.format([_model.number]));
    } on ConflictError catch (e) {
      _conflict(e);
    } on ValidationError catch (e) {
      _snackBar(e.message);
      _toggleProcessing(false);
    } catch (e) {
      _snackBar(AppLocalizations.of(context)!.err);
      _toggleProcessing(false);
    }
  }

  void _snackBar(String message) =>
      ScaffoldMessenger.of(context).showSnackBar(SnackBar(
        content: Text(message),
      ));

  void _onSave() async {
    _toggleProcessing(true);
    try {
      final model = await mailbox.save(_model.id, _model.version, _model.data);
      _model.update(model);
      _composeState.setPristine();
      _snackBar(AppLocalizations.of(context)!.saved);
    } on ConflictError catch (e) {
      _conflict(e);
    } on ValidationError catch (e) {
      _snackBar(e.message);
    } catch (e) {
      _snackBar(AppLocalizations.of(context)!.err);
    } finally {
      _toggleProcessing(false);
    }
  }

  void _conflict(ConflictError e) {
    ScaffoldMessenger.of(context).showSnackBar(SnackBar(
      content: Text(AppLocalizations.of(context)!.err),
      action: SnackBarAction(
        label: AppLocalizations.of(context)!.ok,
        onPressed: () {
          _model.update(e.data);
          _composeState.reset();
        },
      ),
    ));
    _toggleProcessing(false);
  }

  void _toggleProcessing(bool val) => setState(() => _processing = val);

  ///
  void _upload() async {
    try {
      final result = await FilePicker.platform.pickFiles(allowMultiple: true);
      _toggleProcessing(true);
      if (result != null) {
        final files = result.paths.map((path) => File(path!));
        final model = await mailbox.upload(_model.id, _model.version, files);
        _model.attachments.addAll(model);
        _composeState.setPristine();
        setState(() {});
      }
    } catch (e) {
      _snackBar(AppLocalizations.of(context)!.err);
    } finally {
      _toggleProcessing(false);
    }
  }

  void _discard() async {
    final ok = await showDialog<bool?>(
      context: context,
      builder: (context) => AlertDialog(
        content: Text(AppLocalizations.of(context)!.discardAsk),
        actions: [
          TextButton(
            child: Text(AppLocalizations.of(context)!.cancel),
            onPressed: () => Navigator.pop(context),
          ),
          TextButton(
            child: Text(AppLocalizations.of(context)!.discard),
            onPressed: () => Navigator.pop(context, true),
          )
        ],
      ),
    );
    if (ok != true) return;
    _toggleProcessing(true);
    try {
      await mailbox.discard(_model.id);
      SchedulerBinding.instance!.addPostFrameCallback((_) {
        Navigator.pop(context);
      });
    } catch (e) {
      _snackBar(AppLocalizations.of(context)!.err);
      _toggleProcessing(false);
    }
  }

  @override
  void dispose() {
    _composeState.dispose();
    super.dispose();
  }
}

class ActionDialog {
  static Future<Map<String, dynamic>?> show(
      BuildContext context, ActionInfo action) {
    var args = <String, dynamic>{};
    return showDialog<Map<String, dynamic>>(
      context: context,
      builder: (context) {
        return AlertDialog(
          title: Text(action.title),
          content: Fields(
              fields: action.arguments,
              onChanged: (values) => args.addAll(values)),
          actions: [
            TextButton(
              child: Text(AppLocalizations.of(context)!.cancel),
              onPressed: () => Navigator.pop(context),
            ),
            TextButton(
              child: Text(AppLocalizations.of(context)!.oK),
              onPressed: () => Navigator.pop(context, args),
            )
          ],
        );
      },
    );
  }
}
